import logging
import os
import uuid
from pathlib import Path
from typing import Dict, Optional, Union

import yaml

from taskcat._cfn.template import Template
from taskcat._client_factory import Boto3Cache
from taskcat._dataclasses import (
    BaseConfig,
    RegionObj,
    S3BucketObj,
    Tag,
    TestObj,
    TestRegion,
    generate_bucket_name,
    generate_regional_bucket_name,
)
from taskcat._legacy_config import legacy_overrides, parse_legacy_config
from taskcat._template_params import ParamGen
from taskcat.exceptions import TaskCatException

LOG = logging.getLogger(__name__)

GENERAL = Path("~/.taskcat.yml").expanduser().resolve()
PROJECT = Path("./.taskcat.yml").resolve()
PROJECT_ROOT = Path("./").resolve()
OVERRIDES = Path("./.taskcat_overrides.yml").resolve()

DEFAULTS = {
    "project": {
        "s3_enable_sig_v2": False,
        "build_submodules": True,
        "package_lambda": True,
        "lambda_zip_path": "lambda_functions/packages",
        "lambda_source_path": "lambda_functions/source",
        "shorten_stack_name": False,
    }
}


class Config:
    def __init__(self, sources: list, uid: uuid.UUID):
        self.config = BaseConfig.from_dict(DEFAULTS)
        self.config.set_source("TASKCAT_DEFAULT")
        self.uid = uid
        for source in sources:
            config_dict: dict = source["config"]
            source_name: str = source["source"]
            source_config = BaseConfig.from_dict(config_dict)
            source_config.set_source(source_name)
            self.config = BaseConfig.merge(self.config, source_config)

    @classmethod
    # pylint: disable=too-many-locals
    def create(
        cls,
        template_file: Optional[Path] = None,
        args: Optional[dict] = None,
        global_config_path: Path = GENERAL,
        project_config_path: Path = PROJECT,
        overrides_path: Path = OVERRIDES,
        env_vars: Optional[dict] = None,
        project_root: Path = PROJECT_ROOT,
        uid: uuid.UUID = None,
    ) -> "Config":
        uid = uid if uid else uuid.uuid4()
        project_source = cls._get_project_source(
            cls, project_config_path, project_root, template_file
        )

        # general
        legacy_overrides(
            Path("~/.aws/taskcat_global_override.json").expanduser().resolve(),
            global_config_path,
            "global",
        )
        sources = [
            {
                "source": str(global_config_path),
                "config": cls._dict_from_file(global_config_path),
            }
        ]

        # project config file
        if project_source:
            sources.append(project_source)

        # template file
        if isinstance(template_file, Path):
            sources.append(
                {
                    "source": str(template_file),
                    "config": cls._dict_from_template(template_file),
                }
            )

        # override file
        legacy_overrides(
            project_root / "ci/taskcat_project_override.json", overrides_path, "project"
        )
        if overrides_path.is_file():
            overrides = BaseConfig().to_dict()
            with open(str(overrides_path), "r") as file_handle:
                override_params = yaml.safe_load(file_handle)
            overrides["project"]["parameters"] = override_params
            sources.append({"source": str(overrides_path), "config": overrides})

        # environment variables
        sources.append(
            {
                "source": "EnvoronmentVariable",
                "config": cls._dict_from_env_vars(env_vars),
            }
        )

        # cli arguments
        if args:
            sources.append({"source": "CliArgument", "config": args})
        return cls(sources=sources, uid=uid)

    # pylint: disable=protected-access
    @staticmethod
    def _get_project_source(base_cls, project_config_path, project_root, template_file):
        try:
            return {
                "source": str(project_config_path),
                "config": base_cls._dict_from_file(project_config_path, fail_ok=False),
            }
        except Exception as e:  # pylint: disable=broad-except
            error = e
            try:
                legacy_conf = parse_legacy_config(project_root)
                return {
                    "source": str(project_root / "ci/taskcat.yml"),
                    "config": legacy_conf.to_dict(),
                }
            except Exception as e:  # pylint: disable=broad-except
                LOG.debug(str(e), exc_info=True)
                if not template_file:
                    raise error

    @staticmethod
    def _dict_from_file(file_path: Path, fail_ok=True) -> dict:
        config_dict = BaseConfig().to_dict()
        if not file_path.is_file() and fail_ok:
            return config_dict
        try:
            with open(str(file_path), "r") as file_handle:
                config_dict = yaml.safe_load(file_handle)
            return config_dict
        except Exception as e:  # pylint: disable=broad-except
            LOG.warning(f"failed to load config from {file_path}")
            LOG.debug(str(e), exc_info=True)
            if not fail_ok:
                raise e
        return config_dict

    @staticmethod
    def _dict_from_template(file_path: Path) -> dict:
        relative_path = str(file_path.relative_to(PROJECT_ROOT))
        config_dict = (
            BaseConfig()
            .from_dict(
                {"project": {"template": relative_path}, "tests": {"default": {}}}
            )
            .to_dict()
        )
        if not file_path.is_file():
            raise TaskCatException(f"invalid template path {file_path}")
        try:
            template = Template(str(file_path)).template
        except Exception as e:
            LOG.warning(f"failed to load template from {file_path}")
            LOG.debug(str(e), exc_info=True)
            raise e
        if not template.get("Metadata"):
            return config_dict
        if not template["Metadata"].get("taskcat"):
            return config_dict
        template_config_dict = template["Metadata"]["taskcat"]
        if not template_config_dict.get("project"):
            template_config_dict["project"] = {}
        template_config_dict["project"]["template"] = relative_path
        if not template_config_dict.get("tests"):
            template_config_dict["tests"] = {"default": {}}
        return template_config_dict

    # pylint: disable=protected-access
    @staticmethod
    def _dict_from_env_vars(
        env_vars: Optional[Union[os._Environ, Dict[str, str]]] = None
    ):
        if env_vars is None:
            env_vars = os.environ
        config_dict: Dict[str, Dict[str, Union[str, bool, int]]] = {}
        for key, value in env_vars.items():
            if key.startswith("TASKCAT_"):
                key = key[8:].lower()
                sub_key = None
                key_section = None
                for section in ["general", "project", "tests"]:
                    if key.startswith(section):
                        sub_key = key[len(section) + 1 :]
                        key_section = section
                if isinstance(sub_key, str) and isinstance(key_section, str):
                    if value.isnumeric():
                        value = int(value)
                    elif value.lower() in ["true", "false"]:
                        value = value.lower() == "true"
                    if not config_dict.get(key_section):
                        config_dict[key_section] = {}
                    config_dict[key_section][sub_key] = value
        return config_dict

    def get_regions(self, boto3_cache: Boto3Cache = None):
        if boto3_cache is None:
            boto3_cache = Boto3Cache()

        region_objects: Dict[str, Dict[str, RegionObj]] = {}
        for test_name, test in self.config.tests.items():
            region_objects[test_name] = {}
            for region in test.regions:
                profile = (
                    test.auth.get(region, test.auth.get("default", "default"))
                    if test.auth
                    else "default"
                )
                region_objects[test_name][region] = RegionObj(
                    name=region,
                    account_id=boto3_cache.account_id(profile),
                    partition=boto3_cache.partition(profile),
                    profile=profile,
                    _boto3_cache=boto3_cache,
                    taskcat_id=self.uid,
                )
        return region_objects

    def get_buckets(self, boto3_cache: Boto3Cache = None):
        regions = self.get_regions(boto3_cache)
        bucket_objects: Dict[str, S3BucketObj] = {}
        bucket_mappings: Dict[str, Dict[str, S3BucketObj]] = {}
        for test_name, test in self.config.tests.items():
            bucket_mappings[test_name] = {}
            for region_name, region in regions[test_name].items():
                if test.s3_regional_buckets:
                    bucket_obj = self._create_regional_bucket_obj(
                        bucket_objects, region, test
                    )
                    bucket_objects[f"{region.account_id}{region.name}"] = bucket_obj
                else:
                    bucket_obj = self._create_legacy_bucket_obj(
                        bucket_objects, region, test
                    )
                    bucket_objects[region.account_id] = bucket_obj
                bucket_mappings[test_name][region_name] = bucket_obj

        return bucket_mappings

    def _create_legacy_bucket_obj(self, bucket_objects, region, test):
        new = False
        object_acl = (
            self.config.project.s3_object_acl
            if self.config.project.s3_object_acl
            else "private"
        )
        sigv4 = not self.config.project.s3_enable_sig_v2
        if not test.s3_bucket and not bucket_objects.get(region.account_id):
            name = generate_bucket_name(self.config.project.name)
            auto_generated = True
            new = True
        elif bucket_objects.get(region.account_id):
            name = bucket_objects[region.account_id].name
            auto_generated = bucket_objects[region.account_id].auto_generated
        else:
            name = test.s3_bucket
            auto_generated = False
        bucket_region = self._get_bucket_region_for_partition(region.partition)
        bucket_obj = S3BucketObj(
            name=name,
            region=bucket_region,
            account_id=region.account_id,
            s3_client=region.session.client("s3", region_name=bucket_region),
            auto_generated=auto_generated,
            object_acl=object_acl,
            sigv4=sigv4,
            taskcat_id=self.uid,
            partition=region.partition,
            regional_buckets=test.s3_regional_buckets,
        )
        if new:
            bucket_obj.create()
        return bucket_obj

    def _create_regional_bucket_obj(self, bucket_objects, region, test):
        _bucket_obj_key = f"{region.account_id}{region.name}"
        new = False
        object_acl = (
            self.config.project.s3_object_acl
            if self.config.project.s3_object_acl
            else "private"
        )
        sigv4 = not self.config.project.s3_enable_sig_v2
        if not test.s3_bucket and not bucket_objects.get(_bucket_obj_key):
            name = generate_regional_bucket_name(region)
            auto_generated = True
            new = True
        elif bucket_objects.get(_bucket_obj_key):
            name = bucket_objects[_bucket_obj_key].name
            auto_generated = bucket_objects[_bucket_obj_key].auto_generated
        else:
            name = test.s3_bucket
            auto_generated = False
        bucket_obj = S3BucketObj(
            name=name,
            region=region.name,
            account_id=region.account_id,
            s3_client=region.session.client("s3", region_name=region.name),
            auto_generated=auto_generated,
            object_acl=object_acl,
            sigv4=sigv4,
            taskcat_id=self.uid,
            partition=region.partition,
            regional_buckets=test.s3_regional_buckets,
        )
        if new:
            bucket_obj.create()
        return bucket_obj

    @staticmethod
    def _get_bucket_region_for_partition(partition):
        region = "us-east-1"
        if partition == "aws-us-gov":
            region = "us-gov-east-1"
        elif partition == "aws-cn":
            region = "cn-north-1"
        return region

    def get_rendered_parameters(self, bucket_objects, region_objects, template_objects):
        parameters = {}
        template_params = self.get_params_from_templates(template_objects)
        for test_name, test in self.config.tests.items():
            parameters[test_name] = {}
            for region_name in test.regions:
                region_params = template_params[test_name].copy()
                for param_key, param_value in test.parameters.items():
                    if param_key in region_params:
                        region_params[param_key] = param_value
                region = region_objects[test_name][region_name]
                s3bucket = bucket_objects[test_name][region_name]
                parameters[test_name][region_name] = ParamGen(
                    region_params, s3bucket.name, region.name, region.client
                ).results
        return parameters

    @staticmethod
    def get_params_from_templates(template_objects):
        parameters = {}
        for test_name, template in template_objects.items():
            parameters[test_name] = template.parameters()
        return parameters

    def get_templates(self, project_root: Path):
        templates = {}
        for test_name, test in self.config.tests.items():
            templates[test_name] = Template(
                template_path=project_root / test.template,
                project_root=project_root,
                s3_key_prefix=f"{self.config.project.name}/",
            )
        return templates

    def get_tests(self, project_root, templates, regions, buckets, parameters):
        tests = {}
        for test_name, test in self.config.tests.items():
            region_list = []
            tag_list = []
            if test.tags:
                for tag_key, tag_value in test.tags.items():
                    tag_list.append(Tag({"Key": tag_key, "Value": tag_value}))
            for region_obj in regions[test_name].values():
                region_list.append(
                    TestRegion.from_region_obj(
                        region_obj,
                        buckets[test_name][region_obj.name],
                        parameters[test_name][region_obj.name],
                    )
                )
            tests[test_name] = TestObj(
                name=test_name,
                template_path=project_root / test.template,
                template=templates[test_name],
                project_root=project_root,
                regions=region_list,
                tags=tag_list,
            )
        return tests



import logging
import time
from pathlib import Path

import requests

import yattag

from ._cfn.stack import Stack
from ._cfn.threaded import Stacker

LOG = logging.getLogger(__name__)


class ReportBuilder:
    """
    This class generates the test report.

    """

    def __init__(self, stacks: Stacker, output_file: Path, version: str = "0.9.0"):
        self._stacks = stacks
        self._output_file = output_file
        self._version = version

    # TODO: refactor for readability
    def generate_report(  # noqa: C901
        self,
    ):  # pylint: disable=too-many-locals, too-many-statements
        doc = yattag.Doc()

        # Type of cfn log return cfn log file
        def get_output_file(region, stack_name, resource_type):
            extension = ".txt"
            if resource_type == "cfnlog":
                location = "{}-{}-{}{}".format(stack_name, region, "cfnlogs", extension)
                return str(location)
            if resource_type == "resource_log":
                location = "{}-{}-{}{}".format(
                    stack_name, region, "resources", extension
                )
                return str(location)
            return None

        def get_teststate(stack: Stack):
            rstatus = stack.status
            if rstatus == "CREATE_COMPLETE":
                status_css = "class=test-green"
            elif rstatus == "CREATE_FAILED":
                status_css = "class=test-red"
            else:
                status_css = "class=test-red"
            return rstatus, status_css

        tag = doc.tag
        text = doc.text
        logo = "taskcat"
        repo_link = "https://github.com/aws-quickstart/taskcat"
        css_url = (
            "https://raw.githubusercontent.com/aws-quickstart/taskcat/master/"
            "assets/css/taskcat_reporting.css"
        )
        output_css = requests.get(css_url).text
        doc_link = "http://taskcat.io"

        with tag("html"):
            with tag("head"):
                doc.stag("meta", charset="utf-8")
                doc.stag("meta", name="viewport", content="width=device-width")
                with tag("style", type="text/css"):
                    text(output_css)
                with tag("title"):
                    text("TaskCat Report")

            with tag("body"):
                tested_on = time.strftime("%A - %b,%d,%Y @ %H:%M:%S")

                with tag("table", "class=header-table-fill"):
                    with tag("tbody"):
                        with tag("th", "colspan=2"):
                            with tag("tr"):
                                with tag("td"):
                                    with tag("a", href=repo_link):
                                        text("GitHub Repo: ")
                                        text(repo_link)
                                        doc.stag("br")
                                    with tag("a", href=doc_link):
                                        text("Documentation: ")
                                        text(doc_link)
                                        doc.stag("br")
                                    text("Tested on: ")
                                    text(tested_on)
                                with tag("td", "class=taskcat-logo"):
                                    with tag("h3"):
                                        text(logo)
            doc.stag("p")
            with tag("table", "class=table-fill"):
                with tag("tbody"):
                    with tag("thread"):
                        with tag("tr"):
                            with tag("th", "class=text-center", "width=25%"):
                                text("Test Name")
                            with tag("th", "class=text-left", "width=10%"):
                                text("Tested Region")
                            with tag("th", "class=text-left", "width=30%"):
                                text("Stack Name")
                            with tag("th", "class=text-left", "width=20%"):
                                text("Tested Results")
                            with tag("th", "class=text-left", "width=15%"):
                                text("Test Logs")

                            for stack in self._stacks.stacks:
                                with tag("tr", "class= test-footer"):
                                    with tag("td", "colspan=5"):
                                        text("")

                                testname = stack.test_name
                                LOG.info(f"Reporting on {str(stack.id)}")
                                status, css = get_teststate(stack)

                                with tag("tr"):
                                    with tag("td", "class=test-info"):
                                        with tag("h3"):
                                            text(testname)
                                    with tag("td", "class=text-left"):
                                        text(stack.region_name)
                                    with tag("td", "class=text-left"):
                                        text(stack.name)
                                    with tag("td", css):
                                        text(str(status))
                                    with tag("td", "class=text-left"):
                                        clog = get_output_file(
                                            stack.region_name, stack.name, "cfnlog"
                                        )
                                        with tag("a", href=clog):
                                            text("View Logs ")
                            with tag("tr", "class= test-footer"):
                                with tag("td", "colspan=5"):
                                    vtag = "Generated by {} {}".format(
                                        "taskcat", self._version
                                    )
                                    text(vtag)

                        doc.stag("p")

            html_output = yattag.indent(
                doc.getvalue(), indentation="    ", newline="\r\n", indent_text=True
            )

            file = open(str(self._output_file.resolve()), "w")
            file.write(html_output)
            file.close()

            return html_output



import logging
import uuid
from functools import partial
from multiprocessing.dummy import Pool as ThreadPool
from typing import Dict, List

import boto3

from taskcat._cfn.stack import Stack, Stacks, StackStatus
from taskcat._client_factory import Boto3Cache
from taskcat._common_utils import merge_dicts
from taskcat._dataclasses import Tag, TestObj, TestRegion
from taskcat.exceptions import TaskCatException

LOG = logging.getLogger(__name__)


def fan_out(func, partial_kwargs, payload, threads):
    pool = ThreadPool(threads)
    if partial_kwargs:
        func = partial(func, **partial_kwargs)
    results = pool.map(func, payload)
    pool.close()
    pool.join()
    return results


class Stacker:

    NULL_UUID = uuid.UUID(int=0)

    def __init__(
        self,
        project_name: str,
        tests: Dict[str, TestObj],
        uid: uuid.UUID = NULL_UUID,
        stack_name_prefix: str = "tCaT",
        shorten_stack_name: bool = False,
        tags: list = None,
    ):
        self.tests = tests
        self.project_name = project_name
        self.stack_name_prefix = stack_name_prefix
        self.shorten_stack_name = shorten_stack_name
        self.tags = tags if tags else []
        self.uid = uuid.uuid4() if uid == Stacker.NULL_UUID else uid
        self.stacks: Stacks = Stacks()

    @staticmethod
    def _tests_to_list(tests: Dict[str, TestObj]):
        return list(tests.values())

    def create_stacks(self, threads: int = 8):
        if self.stacks:
            raise TaskCatException("Stacker already initialised with stack objects")
        tests = self._tests_to_list(self.tests)
        tags = [Tag({"Key": "taskcat-id", "Value": self.uid.hex})]
        tags += [
            Tag(t)
            for t in self.tags
            if t.key not in ["taskcat-project-name", "taskcat-test-name", "taskcat-id"]
        ]
        fan_out(self._create_stacks_for_test, {"tags": tags}, tests, threads)

    def _get_stack_name_for_test(self, test):
        prefix = f"{self.stack_name_prefix}-" if self.stack_name_prefix else ""
        if self.shorten_stack_name:
            return "{}{}-{}".format(prefix, test.name, self.uid.hex[:6])
        return "{}{}-{}-{}".format(prefix, self.project_name, test.name, self.uid.hex)

    def _create_stacks_for_test(self, test, tags, threads: int = 32):
        stack_name = self._get_stack_name_for_test(test)
        tags.append(Tag({"Key": "taskcat-project-name", "Value": self.project_name}))
        tags.append(Tag({"Key": "taskcat-test-name", "Value": test.name}))
        tags += test.tags
        partial_kwargs = {
            "stack_name": stack_name,
            "template": test.template,
            "tags": tags,
            "test_name": test.name,
        }
        stacks = fan_out(Stack.create, partial_kwargs, test.regions, threads)
        self.stacks += stacks

    # Not used by tCat at present
    def update_stacks(self):
        raise NotImplementedError()

    def delete_stacks(self, criteria: dict = None, deep=False, threads=32):
        if deep:
            raise NotImplementedError("deep delete not yet implemented")
        fan_out(
            self._delete_stacks_per_client,
            None,
            self._group_stacks(self.stacks.filter(criteria)),
            threads,
        )

    def _delete_stacks_per_client(self, stacks, threads=8):
        fan_out(self._delete_stack, None, stacks["Stacks"], threads)

    @staticmethod
    def _delete_stack(stack: Stack):
        stack.delete(stack_id=stack.id, client=stack.client)
        stack.refresh()

    def status(self, recurse: bool = False, threads: int = 32, **kwargs):
        if recurse:
            raise NotImplementedError("recurse not implemented")
        stacks = self.stacks.filter(kwargs)
        per_region_stacks = self._group_stacks(stacks)
        results = fan_out(self._status_per_client, None, per_region_stacks, threads)
        statuses: Dict[str, dict] = {"IN_PROGRESS": {}, "COMPLETE": {}, "FAILED": {}}
        for region in results:
            for status in region:
                statuses[status[1]][status[0]] = status[2]
        return statuses

    def _status_per_client(self, stacks, threads: int = 8):
        return fan_out(self._status, None, stacks["Stacks"], threads)

    @staticmethod
    def _status(stack: Stack):
        for status_group in ["COMPLETE", "IN_PROGRESS", "FAILED"]:
            if stack.status in getattr(StackStatus, status_group):
                return stack.id, status_group, stack.status_reason
        raise TaskCatException(f"Invalid stack {stack}")

    def events(self, recurse=False, threads: int = 32, **kwargs):
        if recurse:
            raise NotImplementedError("recurse not implemented")
        per_region_stacks = self._group_stacks(self.stacks)
        results = fan_out(
            self._events_per_client, {"criteria": kwargs}, per_region_stacks, threads
        )
        return merge_dicts(results)

    def _events_per_client(self, stacks, criteria, threads: int = 8):
        results = fan_out(
            self._describe_stack_events,
            {"criteria": criteria},
            stacks["Stacks"],
            threads,
        )
        return merge_dicts(results)

    @staticmethod
    def _describe_stack_events(stack: Stack, criteria):
        return {stack.id: stack.events().filter(criteria)}

    def resources(self, recurse=False, threads: int = 32, **kwargs):
        if recurse:
            raise NotImplementedError("recurse not implemented")
        results = fan_out(
            self._resources_per_client,
            {"criteria": kwargs},
            self._group_stacks(self.stacks),
            threads,
        )
        return merge_dicts(results)

    def _resources_per_client(self, stacks, criteria, threads: int = 8):
        results = fan_out(
            self._resources, {"criteria": criteria}, stacks["Stacks"], threads
        )
        return merge_dicts(results)

    @staticmethod
    def _resources(stack: Stack, criteria):
        return {stack.id: stack.resources().filter(criteria)}

    @classmethod
    def from_existing(
        cls,
        uid: uuid.UUID,
        project_name: str,
        tests: Dict[str, TestObj],
        include_deleted=False,
        recurse=False,
        threads=32,
    ):
        if include_deleted:
            raise NotImplementedError("including deleted stacks not implemented")
        if recurse:
            raise NotImplementedError("recurse not implemented")
        clients: Dict[boto3.client, List[TestRegion]] = {}
        for test in tests.values():
            for region in test.regions:
                client = region.client("cloudformation")
                if client not in clients:
                    clients[client] = []
                clients[client].append(region)
        results = fan_out(
            Stacker._import_stacks_per_client,
            {"uid": uid, "project_name": project_name, "tests": tests},
            clients.items(),
            threads,
        )
        stacker = Stacker(project_name, tests, uid)
        stacker.stacks = Stacks([item for sublist in results for item in sublist])
        return stacker

    @staticmethod
    def _import_stacks_per_client(clients, uid, project_name, tests):
        # pylint: disable=too-many-locals
        stacks = Stacks()
        client, region = clients
        for page in client.get_paginator("describe_stacks").paginate():
            for stack_props in page["Stacks"]:
                if stack_props.get("ParentId"):
                    continue
                match = False
                project = ""
                test = ""
                for tag in stack_props["Tags"]:
                    k, v = (tag["Key"], tag["Value"])
                    if k == "taskcat-id" and v == uid.hex:
                        match = True
                    elif k == "taskcat-test-name" and v in tests:
                        test = v
                    elif k == "taskcat-project-name" and v == project_name:
                        project = v
                if match and test and project:
                    stack = Stack.import_existing(
                        stack_props, tests[test].template, region[0], test, uid
                    )
                    stacks.append(stack)
        return stacks

    @staticmethod
    def _group_stacks(stacks: Stacks) -> List[dict]:
        stacks_by_client: dict = {}
        for stack in stacks:
            client = stack.client
            if client not in stacks_by_client:
                stacks_by_client[client] = {"Client": client, "Stacks": []}
            stacks_by_client[client]["Stacks"].append(stack)
        return [stacks_by_client[r] for r in stacks_by_client]

    @staticmethod
    def list_stacks(profiles, regions):
        stacks = fan_out(
            Stacker._list_per_profile,
            {"regions": regions, "boto_cache": Boto3Cache()},
            profiles,
            threads=8,
        )
        return [stack for sublist in stacks for stack in sublist]

    @staticmethod
    def _list_per_profile(profile, regions, boto_cache):
        stacks = fan_out(
            Stacker._get_taskcat_stacks,
            {"boto_cache": boto_cache, "profile": profile},
            regions,
            threads=len(regions),
        )
        return [stack for sublist in stacks for stack in sublist]

    @staticmethod
    def _get_taskcat_stacks(region, boto_cache: Boto3Cache, profile: str):
        stacks = []
        try:
            cfn = boto_cache.client("cloudformation", profile=profile, region=region)
            for page in cfn.get_paginator("describe_stacks").paginate():
                for stack_props in page["Stacks"]:
                    if stack_props.get("ParentId"):
                        continue
                    stack_id = stack_props["StackId"]
                    stack_name = stack_id.split("/")[1]
                    stack = {
                        "region": region,
                        "profile": profile,
                        "stack-id": stack_id,
                        "stack-name": stack_name,
                    }
                    for tag in stack_props["Tags"]:
                        k, v = (tag["Key"], tag["Value"])
                        if k.startswith("taskcat-"):
                            stack[k] = v
                    if stack.get("taskcat-id"):
                        stack["taskcat-id"] = uuid.UUID(stack["taskcat-id"])
                        stacks.append(stack)
        except Exception as e:  # pylint: disable=broad-except
            LOG.warning(
                f"Failed to fetch stacks for region {region} using profile "
                f"{profile} {type(e)} {e}"
            )
            LOG.debug(f"Traceback:", exc_info=True)
        return stacks



import logging
import time

from reprint import output
from taskcat._cfn.threaded import Stacker as TaskcatStacker
from taskcat._logger import PrintMsg

LOG = logging.getLogger(__name__)


class TerminalPrinter:
    def __init__(self):
        self._buffer_type = "list"
        self.buffer = self._add_buffer()

    def _add_buffer(self):
        with output(output_type=self._buffer_type) as output_buffer:
            return output_buffer

    def report_test_progress(self, stacker: TaskcatStacker, poll_interval=10):
        _status_dict = stacker.status()
        while self._is_test_in_progress(_status_dict):
            for stack in stacker.stacks:
                self._print_stack_tree(stack, buffer=self.buffer)
            time.sleep(poll_interval)
            self.buffer.clear()
            _status_dict = stacker.status()

        self._display_final_status(stacker)

    @staticmethod
    def _print_stack_tree(stack, buffer):
        padding_1 = "         "
        buffer.append(
            "{}{}stack {} {}".format(padding_1, "\u250f ", "\u24c2", stack.name)
        )
        if stack.descendants():
            for nested_stack in stack.descendants():
                buffer.append(
                    "{}{}stack {} {}".format(
                        padding_1, "\u2523 ", "\u24c3", nested_stack.name
                    )
                )
        buffer.append("{}{} region: {}".format(padding_1, "\u2523", stack.region_name))
        buffer.append(
            "{}{}status: {}{}{}".format(
                padding_1, "\u2517 ", PrintMsg.white, stack.status, PrintMsg.rst_color
            )
        )

    @staticmethod
    def _display_final_status(stacker: TaskcatStacker):
        for final_stack in stacker.stacks:
            LOG.info("{}stack {} {}".format("\u250f ", "\u24c2", final_stack.name))
            if final_stack.descendants():
                for nested_stack in final_stack.descendants():
                    LOG.info(
                        "{}stack {} {}".format("\u2523 ", "\u24c3", nested_stack.name)
                    )
            LOG.info("{} region: {}".format("\u2523", final_stack.region_name))

            LOG.info(
                "{}status: {}{} {}".format(
                    "\u2517 ", PrintMsg.white, final_stack.status, PrintMsg.rst_color
                )
            )

    @staticmethod
    def _is_test_in_progress(status_dict, status_condition="IN_PROGRESS"):
        return bool(len(status_dict[status_condition]) > 0)
